HMMNASE <-
function(allecounts, dist, reps=1, epsilon0=0.075,maxiter=20,cnv=NULL, dist.dep=FALSE)
  {
    # Inputs:
    # 1. allecounts:           a (#SNP * 2 * #animal) array. Element [s, 1, a] is the alle count
    #                          of the first allel (A) for SNP S of animal a; and element [s, 2, a] is the alle 
    #                          count of the second allel (T) for SNP S of animal a.
    # 2. reps:                 Number of replications. reps=1 for real data set.
    # 3. epsilon0:             error rate. Between 0 and 1
    # 8. maxiter:              maximum number of iteration
    # 9. cnv:                  convergence criterion. HMM function stops with convergence if the number of estimated 
    #                          genotype state is different with less than T*n*cnv from the last iteration. 
    #                          If cnv=NULL, then the default value cnv=0.01. 
    # Outputs/Values:
    # 1. allecounts:           3 dimentional data matrix with RNA read counts [SNP, Allele, Animal].
    # 2. GenoEsti:             3 dimentional matrix with genotype prediction result [SNP, Animal, replicate].
    #                          For real data, #replicate=1, so the dimention is [#SNPs, #Animals, 1].
    # 3. Parameters:           Estimation of paramter values after convergence with length 30.
    #                          c(epsilon0,alpha10,beta010,alpha20,beta020,as.vector(Amat))
    #                          Amat is a 5 by 5 transition matrix.
    
    # output_file<-paste(wd, "NASE_NDD_output_", idid, ".txt", sep="")
    # cat(idid, "\n")
    T1<-dim(allecounts)[1]
    n<-dim(allecounts)[3]
    
    MacMin<-.Machine$double.xmin
    
    neg.loc<-which(allecounts<0,arr.ind=TRUE)
    if (!is.null(neg.loc))
    { # cat("Some counts are less than 0 and the values were changed to be 0","\n", file=output_file, append=TRUE)
      cat("Some counts are less than 0 and the values were changed to be 0","\n") 
      allecounts[neg.loc]=0}
    
    GenotypeEsti<-array(0,c(T1,n,reps))
    Parameters<-NULL
    deltas<-array(0,c(T1,n,reps))
    if(is.null(cnv)){
      cnv<-T1*n*0.01
    } else {
      cnv<-T1*n*cnv
    }
    
    for (W in 1:reps)
    {
      Logfx<-function(gt,X,epsilon)
      {
        M<-length(gt)
        T1<-dim(X)[1]
        n<-dim(X)[3]
        logfx<-array(0,c(T1,n,M))
        nt<-apply(X,c(1,3),sum)
        for (i in 1:M)  
        {
          if (gt[i]==1) ## AA
          { pt<-1-epsilon
            logfx[,,i]<-matrix(unlist(lapply(dbinom(X[,1,],size=nt, prob=pt)+MacMin,eln)),T1,n)}
          if (gt[i]==2) ## AT-NASE
          { pt<-0.5
            logfx[,,i]<-matrix(unlist(lapply(dbinom(X[,1,],size=nt, prob=pt)+MacMin,eln)),T1,n)}
          if (gt[i]==3) ## TT
          { pt<-epsilon
            logfx[,,i]<-matrix(unlist(lapply(dbinom(X[,1,],size=nt, prob=pt)+MacMin,eln)),T1,n)}  
        }  
        return(logfx)  
      }
      
      M<-3
      
      Amat<-array(1/M,c(M,M,T1-1))
      if (dist.dep==TRUE)
      {
        a_mat<-matrix(1/M, M, M)
        rho<-10
        Amat<-Amat_fun(a_mat, rho=rho, dist, T1, M) 
      }
      
      maxlocation<-matrix(0,T1,n)
      ObsLikelihood<-0
      
      repcounts<-1
      converge<-TRUE
      
      repeat
      {
        if(T1<2){
          warning ("too few SNP")
          converge<-FALSE
          break
        }
        
        ## step 1: computing eln(alpha), eln(beta), and eln(Px)
        
        pi0<-rep(1/M,M)
        
        ## compute eln(alpha)
        
        lnalphas<-array(0,c(T1,M,n))
        Logfxmat<-Logfx(c(1:M),allecounts,epsilon0)
        
        for (k in 1:n)
        {
          for (i in 1:M)
          {
            lnalphas[1,i,k]<-elnproduct(eln(pi0[i]),Logfxmat[1,k,i])
          }
          for (t in 2:T1)
          {
            for (j in 1:M)
            { logalpha<--9999
              for (i in 1:M)
              {
                logalpha<-elnsum(logalpha,elnproduct(lnalphas[t-1,i,k],eln(Amat[i,j,t-1])))
              }
              lnalphas[t,j,k]<-elnproduct(logalpha,Logfxmat[t,k,j])
            }
          }
        }
        
        
        ## compute eln(beta)
        
        lnbetas<-array(0,c(T1,M,n))
        for (k in 1:n)
        {
          for (t in (T1-1):1)
          {
            for (i in 1:M)
            { logbeta<--9999
              for (j in 1:M)
              {logbeta<-elnsum(logbeta,elnproduct(eln(Amat[i,j,t]), elnproduct(Logfxmat[t+1,k,j],lnbetas[t+1,j,k])))}
              lnbetas[t,i,k]<-logbeta
            }
          }
        } 
        
        ## compute Lk in log-space
        ## Lk is the posterior probability of the state given the observations.
        
        lnLk<-Lk<-array(0,c(T1,M,n))
        for (k in 1:n)
        {
          for (t in 1:T1)
          {
            normalizer<--9999
            for (i in 1:M)
            {lnLk[t,i,k]<-elnproduct(lnalphas[t,i,k],lnbetas[t,i,k])
             normalizer<-elnsum(normalizer,lnLk[t,i,k])
            }
            for (i in 1:M)
            {lnLk[t,i,k]<-elnratio(lnLk[t,i,k],normalizer)
             Lk[t,i,k]<-eexp(lnLk[t,i,k])
            }
          }
        }
        
        ## compute Hklt in log-space
        
        lnHklt<-Hklt<-array(0,c(T1,M,M,n))
        
        for (k in 1:n)
        {
          for (t in 1:(T1-1))
          {
            normalizer<--9999
            for (i in 1:M)
            { 
              for (j in 1:M)
              { lnHklt[t,i,j,k]<-elnproduct(lnalphas[t,i,k],elnproduct(eln(Amat[i,j,t]),elnproduct(Logfxmat[t+1,k,j],lnbetas[t+1,j,k])))
                normalizer<-elnsum(normalizer,lnHklt[t,i,j,k])
              }
            }
            for (i in 1:M)
            { for (j in 1:M)
            {lnHklt[t,i,j,k]<-elnratio(lnHklt[t,i,j,k],normalizer)
             Hklt[t,i,j,k]<-eexp(lnHklt[t,i,j,k])
            }
            }
          }
        }
        
        
        if (dist.dep==FALSE)
        {
          ## Update parameters akl  
          nAmat1<-matrix(0,M,M) 
          for (i in 1:M)
            for (j in 1:M)
            {
              numerator<--9999
              denominator<--9999
              for (k in 1:n)
              {
                for (t in 1:(T1-1))
                {numerator<-elnsum(numerator,lnHklt[t,i,j,k])
                 denominator<-elnsum(denominator,lnLk[t,i,k])
                }
              } 
              nAmat1[i,j]<-eexp(elnratio(numerator,denominator))
            }
          nAmat<-array(rep(nAmat1,T1-1),c(M, M, T1-1))
        }
        
        if (dist.dep==TRUE)
        {
          ## Update parameters akl and rho
          a_find_mat_func<-function(hx, dist, i, rho, Hklt)
          {            
            Hklt_i <- Hklt[,i,,]
            Hklt_i_pigsum <- apply(Hklt_i,c(1,2),sum)
            #Hklt_i_0<-Hklt_i
            #Hklt_i_0[,,]<-0
            Hklt_k_timesum<-apply(Hklt_i,2,sum)
            trans_Hklt_k<-matrix(rep(0,M*(T1-1)),nrow=(T1-1))
            for (t in 1:(T-1))
            {
              trans_Hklt_k_d <- Hklt_k_timesum*(1-exp(-dist[t]*rho))/hx
              trans_Hklt_k[t,] <- trans_Hklt_k_d
              trans_Hklt_k[t,i] <- 1-sum(trans_Hklt_k_d)
            }
            return(sum(delete0product(Hklt_i_pigsum[1:(T1-1),],log(trans_Hklt_k))))
          }
          
          
          a_find_llimit_func<-function(dist, i, rho, Hklt)
          {
            # This function is to give a lower limit of the searching range of x in using uniroot of
            # function p_find_mat_func. For each i, first run this function, then using uniroot on 
            # function p_find_mat_func for the root x.
            
            Hklt_noi <- Hklt[,i,-i,] # Hklt_noi is an array of (T-1)*(M-1)*n for a fixed i (fix the second dimension).
            M11<-apply(Hklt_noi,2,sum)
            M1<-sum(Hklt[,i,,])
            #M21<-M1*(1-exp(-(posi_vec[2:T]-posi_vec[1:(T-1)])^r))
            max_M111<-max(M1,M11)
            
            return(max_M111+min(max_M111,1)*1e-10)
          }
          
          
          r_find_func <- function(rho, a_mat, dist, Hklt, T1)
          {
            # This function is part of the conditional expectation function in E-M, and it is used 
            # to search an optimal value of r.
            #
            
            trans_mat_r <- Amat_fun(a_mat, rho, dist, T1, M)
            trans_mat_r<-wrap(trans_mat_r,map=list(3,1,2))
            if (T1==2)
            {MATRIX <- delete0product(apply(Hklt[1,,,], c(1,2), sum),log(trans_mat_r[1,,]))}
            if (T1>2)
            {MATRIX <- delete0product(apply(Hklt[1:(T1-1), , , ], c(1,2,3), sum),log(trans_mat_r))}
            return(sum(MATRIX))
          }
          
          a_mat_ud<-matrix(c(rep(0,M*M)),nrow=M)
          for (k in 1: M)
          {
            llimit <- a_find_llimit_func(dist, k, rho, Hklt)
            xvalvec<-matrix(seq(llimit,llimit+5000,length=1000),500,1)
            yvec<-apply(xvalvec,1,a_find_mat_func,dist=dist, i=k, rho=rho, Hklt=Hklt)
            x_vec<-xvalvec[which.max(yvec)]
            # x_vec <- optimize(a_find_mat_func, c(llimit,llimit+500), maximum=TRUE, tol = 0.001, dist=dist, i=i, rho=rho, Hklt=Hklt)[[1]]
            a_mat_ud[k,] <- apply(Hklt[,k,,],2,sum)/x_vec  
            a_mat_ud[k,k] <- 0              
          }   
          
          
          rho_ud<-optimize(r_find_func, c(0, 1), maximum=TRUE, tol = 0.001, a_mat=a_mat_ud, dist=dist, Hklt=Hklt, T1=T1)[[1]] 
          
          # eps_p_r_pi <-max(abs(a_mat_ud - a_mat), abs(rho_ud-rho))
          a_mat <- a_mat_ud # pudate p-mat in the transition matrix
          rho <- rho_ud # update r in the transition matrix  
          
          
          nAmat<-Amat_fun(a_mat, rho, dist, T1, M)
          
        }
        
        Logfxmat0<-wrap(Logfxmat, map=list(1, 3, 2))
        class(Logfxmat0)<-"numeric"
        
        nAmat0<-wrap(nAmat,map=list(3,1,2))
        if (T1>2)
        {nObsLikelihood<-sum(Lk[1,,]*log(matrix(rep(pi0,n),M,n,byrow=T)))+sum(delete0product(apply(Hklt[-T1,,,],c(1,2,3),sum),log(nAmat0)))+sum(Lk*Logfxmat0)}
        if (T1==2)
        {nObsLikelihood<-sum(Lk[1,,]*log(matrix(rep(pi0,n),M,n,byrow=T)))+sum(delete0product(apply(Hklt[1,,,],c(1,2),sum),log(nAmat0[1,,])))+sum(Lk*Logfxmat0)}   
        
        
        nit<-apply(allecounts,c(1,3),sum)
        nepsilon<-sum(Lk[,1,]*(nit-allecounts[,1,])+Lk[,3,]*allecounts[,1,])/sum((Lk[,1,]+Lk[,3,])*nit)
        
        Lkorder<-apply(Lk,c(1,3),order)
        nmaxlocation<-Lkorder[3,,]
        
        cond<-(abs(nObsLikelihood-ObsLikelihood)<1e-3)
        if (!is.na(cond)){
          if ((sum(nmaxlocation!=maxlocation)<cnv)||(repcounts>maxiter)||(abs(nObsLikelihood-ObsLikelihood)<1e-3))
          {
            if (repcounts>maxiter) {
              # cat("Too many iterations. Not converge!","\n", file=output_file, append=TRUE)
              cat("Too many iterations. Not converge!","\n")
              converge<-FALSE
            }
            
            if (sum(nmaxlocation!=maxlocation)<cnv) cat("States not change. Converges.","\n") # cat("States not change. Converges.","\n", file=output_file, append=TRUE)
            if (abs(nObsLikelihood-ObsLikelihood)<1e-3) cat("Likelihood not change. Converges.","\n") # cat("Likelihood not change. Converges.","\n", file=output_file, append=TRUE)
            break
          }} else {
            cat("A mat contains zeroes try changing chunk length","\n")
            # cat("A mat contains zeroes try changing chunk length","\n", file=output_file, append=TRUE)
            converge<-FALSE
            break
          }
        
        cat(W,"replicates",repcounts,"iteration","\n") 
        cat("Not matches", sum(nmaxlocation!=maxlocation),"Likelihood",nObsLikelihood,"\n")
        
        # cat(W,"replicates",repcounts,"iteration","\n", file=output_file, append=TRUE) 
        # cat("Not matches", sum(nmaxlocation!=maxlocation),"Likelihood",nObsLikelihood,"\n", file=output_file, append=TRUE)
        
        Amat<-nAmat
        epsilon0<-nepsilon
        maxlocation<-nmaxlocation
        ObsLikelihood<-nObsLikelihood
        
        repcounts<-repcounts+1 
        cat("Amat=", Amat, "\n")
        cat("epsilon0=",epsilon0,"\n")
#         cat("Amat=", Amat, "\n", file=output_file, append=TRUE)
#         cat("epsilon0=",epsilon0,"\n", file=output_file, append=TRUE)
      }
      
      ## Lkorder<-apply(Lk,c(1,3),order)
      ## maxlocation<-Lkorder[4,,]
      
      ## matches is the vector representing number of locations where genotype prediction are correct.
      
      ##matches[W]<-sum(states[,,W]==maxlocation)
      GenotypeEsti[,,W]<-maxlocation
      #      Parameters[W,]<-c(epsilon0,as.vector(Amat))
      if (dist.dep==FALSE)
      {Parameters<-rbind(Parameters,c(epsilon0,as.vector(Amat)))}
      if (dist.dep==TRUE)
      {Parameters<-rbind(Parameters,c(epsilon0,rho,as.vector(a_mat)))}
      
    }
    
    rsobj<-list(allecounts=allecounts,GenoEsti=GenotypeEsti, Parameters=Parameters,Convergence=converge,niter=repcounts)
    return(rsobj)
  }
